/* @import url("https://cdnjs.cloudflare.com/ajax/libs/jstree/3.3.8/themes/default/style.min.css"); */
@import url("https://unpkg.com/vue-material/dist/vue-material.min.css");
@import url("https://unpkg.com/vue-material/dist/theme/default.css");
/* @import url("https://unpkg.com/vue-material/dist/theme/default-dark.css"); */

@import url("https://code.jquery.com/ui/1.12.1/themes/base/jquery-ui.css");
@import url("https://cdn.jsdelivr.net/gh/bmaranville/treejs@latest/src/index.css");
@import url("https://fonts.googleapis.com/css?family=Roboto:300,400,500,700,400italic|Material+Icons");

@import "./web_reduction_filebrowser.css";
@import "./splitting.css";
@import "./menu.css";
@import "./metadata_table.css";
@import "./params_tree.css";
/* @import "jquery-ui-1.12.1.css"; */
/* @import "jstree-3.3.min.js"; */
