
    .split {
        -webkit-box-sizing: border-box;
        -moz-box-sizing: border-box;
        box-sizing: border-box;
        overflow-y: auto;
        overflow-x: hidden;
    }
    
    .content {
        border: 1px solid #C0C0C0;
        box-shadow: inset 0 1px 2px #e4e4e4;
        background-color: #fff;
    }
    
    .gutter {
        background-color: transparent;
        background-repeat: no-repeat;
        background-position: 50%;
    }
    
    .gutter.gutter-horizontal {
        cursor: col-resize;
        /* background-image: url('../grips/vertical.png'); */
    }
    
    .gutter.gutter-vertical {
        cursor: row-resize;
        background-color: #F6F6F6;
        /* background-image: url('../grips/horizontal.png'); */
    }
    
    .split.split-horizontal,
    .gutter.gutter-horizontal {
        position: relative;
    }
