/* @import "material-design-icons/iconfont/material-icons.css"; */
@import "@mdi/font/css/materialdesignicons.min.css";
@import "vue-material/dist/vue-material.min.css";
@import "vue-material/dist/theme/default.css";
/* @import "roboto-fontface/css/roboto/roboto-fontface.css"; */
/* @import 'material-icons/iconfont/material-icons.scss'; */

@import "@widgetjs/tree/src/index.css";
/* @import "./treejs.css"; */
/* @import "jquery-ui-dist/jquery-ui.min.css"; */
@import "./web_reduction_filebrowser.css";
@import "./splitting.css";
@import "./menu.css";
@import "./metadata_table.css";
@import "./params_tree.css";
/* @import "jquery-ui-1.12.1.css"; */
/* @import "jstree-3.3.min.js"; */
