export const instruments = {};

import dcs from './ncnr.dcs.js';
import ospec from './ncnr.ospec.js';
import refl from './ncnr.refl.js';
import sans from './ncnr.sans.js';
import tas from './ncnr.tas.js';
import vsans from './ncnr.vsans.js';
import usans from './ncnr.usans.js';
import xrr from './xrr.js';

instruments['ncnr.dcs'] = dcs;
instruments['ncnr.ospec'] = ospec;
instruments['ncnr.refl'] = refl;
instruments['ncnr.sans'] = sans;
instruments['ncnr.tas'] = tas;
instruments['ncnr.vsans'] = vsans;
instruments['ncnr.usans'] = usans;
instruments['xrr'] = xrr;