
.dropbtn {
    background-color: #4CAF50;
    color: white;
    padding: 4px;
    font-size: 14px;
    border: none;
    cursor: pointer;
}

.dropdown {
/*    position: relative;*/
    display: inline-block;
    float: right;
    z-index: 111;
}

.dropdown-content {
    display: none;
    position: absolute;
    right: 0;
    background-color: #f9f9f9;
    min-width: 160px;
    box-shadow: 0px 4px 6px 0px rgba(0,0,0,0.2);
    z-index: 111;
}

.dropdown-content span {
    color: black;
    padding: 4px 6px;
    text-decoration: none;
    display: block;
}

.dropdown-content span:hover {background-color: #f1f1f1}

.dropdown:hover .dropdown-content {
    display: block;
}

.dropdown:hover .dropbtn {
    background-color: #3e8e41;
}

#main_menu ul{
/*    float: left;  */
    white-space: nowrap;
    width: auto;
}
#main_menu li {
    padding-left: 2px !important;
}
.md-list.md-dense .md-list-item-content {
    min-height: 24px;
    max-height: 30px;
    font-size: 13px;
    padding-top: 0;
    padding-bottom: 0;
}

.md-list .md-list-item-content {
    min-height: 40px;
    max-height: 52px;
    padding-top: 0;
    padding-bottom: 0;
}

.md-select-menu, .md-select-menu * {
  -webkit-transition: none !important;
  -moz-transition: none !important;
  -o-transition: none !important;
  transition: none !important;
}

.md-field {
    margin-bottom: 0;
}

.md-list-item-expand {
    background-color: whitesmoke;
}