import { FileinfoUi } from './fileinfo.js';
import { FloatUi, IntUi } from './number.js';
import { StrUi } from './string.js';
import { BoolUi } from './boolean.js';
import { OptUi } from './opt.js';
import { ScaleUi } from './scale.js';
import { IndexUi } from './index.js';
import { PatchUi } from './patch.js';
import { CoordinateUi } from './coordinate.js';
import { RangeUi } from './range.js';

export const Components = {
  FileinfoUi,
  FloatUi,
  IntUi,
  StrUi,
  BoolUi,
  OptUi,
  ScaleUi,
  IndexUi,
  "patch_metadata-ui": PatchUi,
  CoordinateUi,
  RangeUi
}